from typing import Callable, Any, TypeAliasType, Iterator
from types import ModuleType
import inspect
import pytest
import time


def has_annotations(f: Callable, param_types: list[object], return_type: object = None) -> bool:
    a = f.__annotations__
    ret = a['return'] if 'return' in a else None
    params = list(a.values())[:-1] if 'return' in a else list(a.values())
    for param, param_type in zip(params, param_types):
        if isinstance(param, TypeAliasType):
            param = param.__value__
        if isinstance(param_type, TypeAliasType):
            param_type = param_type.__value__
        if param != param_type:
            return False
    return ret == return_type


def has_no_annotations(f: Callable) -> bool:
    return len(f.__annotations__) == 0


def imported_modules_of(module: ModuleType) -> Iterator[ModuleType]:
    return map(lambda m: m[1], filter(lambda m: inspect.ismodule(m[1]), inspect.getmembers(module)))


def eidp_test(message: str, minus_points: float, task: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    def wrapper(test: Callable[..., Any]) -> Callable[..., Any]:
        test.msg = message
        test.minus_points = minus_points
        test.is_test = True
        test.task = task
        return test
    return wrapper


class ResultsCollector:
    def __init__(self):
        self.reports = []
        self.collected = 0
        self.exitcode = 0
        self.passed = 0
        self.failed = 0
        self.xfailed = 0
        self.skipped = 0
        self.total_duration = 0

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        outcome = yield
        report = outcome.get_result()
        if report.when == 'call':
            self.reports.append(report)

    def pytest_collection_modifyitems(self, items):
        self.collected = len(items)

    def pytest_terminal_summary(self, terminalreporter, exitstatus):
        self.passed = len(terminalreporter.stats.get('passed', []))
        self.failed = len(terminalreporter.stats.get('failed', []))
        self.xfailed = len(terminalreporter.stats.get('xfailed', []))
        self.skipped = len(terminalreporter.stats.get('skipped', []))
        self.total_duration = time.time() - terminalreporter._sessionstarttime

    def get_results(self) -> tuple[list[str], list[str], list[str]]:
        passed = []
        failed = []
        skipped = []

        for report in self.reports:
            if report.outcome == 'passed':
                passed.append(report.nodeid.split('::')[-1])
            elif report.outcome == 'failed':
                failed.append(report.nodeid.split('::')[-1])
            elif report.outcome == 'skipped':
                skipped.append(report.nodeid.split('::')[-1])

        return passed, failed, skipped
